@@ -12,6 +12,9 @@ from datetime import datetime
 import string
 from StringIO import StringIO
 
+from xlocal import xlocal
+
+
 
 
 # tables
@@ -40,7 +43,9 @@ def success(value):
   
 def failure(reason):
   return None, reason
-  
+ 
+
+
 def value(mval):
   return mval[0]
   
@@ -74,16 +79,29 @@ def chain(*fns):
     return val
     
   return chain_  
-  
+
+def flatten(sequences):
+  """
+  Turn a sequence of sequences into a single sequence
+  [[1,2,3], [4,5,6]] -> [1,2,3,4,5,6]
+  """
+  return success([
+    item
+    for sequence in sequences
+    for item in sequence
+  ])
+
+
 def pipe(val, *fns):
 
   m_val = success(val)
-  for f in fns:
+  for i,f in enumerate(fns):
     try:
       m_val = bind(m_val, f)
     except Exception, e:
+      raise
       # this should never happen
-      m_val = failure(e)
+      m_val = failure("{} errored at step {}:  {}".format(f.__name__, i, e))
   return m_val
 
 
@@ -114,7 +132,13 @@ def select(*columns):
   
 
   def _select(seq):
-    Record = namedtuple(seq.schema.__name__, fields)
+    schema = getattr(seq,'schema',None)
+    if schema:
+      name = schema.__name__
+    else:
+      name = "row"
+
+    Record = namedtuple(name, fields)
     
     return success([
       Record(*[col(item) for col in cols])
@@ -210,24 +234,34 @@ def when(**context):
   _when.invoke = invoke
   return _when
 
-def where(clause):
-  boolean_op = parse(clause)
+def where(clause, **options):
+  boolean_op = parse(clause, **options)
   
-  def _where(seq):
+  def _where(seq, ctx):
     return success([
       item
       for item in seq
-      if boolean_op(item)
+      if boolean_op(item, ctx)
     ])
     
   return _where
     
+def unit(func):
+  "Invoke a function returning (success, error)"
+
+  def _(value):
+    try:
+      return success(func(value))
+    except Exception, e:
+      return failure(e)
+  return _
     
 
 def each(method):
   fmap = functools.partial(map, method)
   def _each(sec):
     return success(fmap(sec))
+
   return _each
 
 
@@ -264,14 +298,22 @@ MULTIPLICATIVE_OPS ={
   '*'  : operator.mul,
   '/'  : operator.div
 }
+SYMBOLS = '+-*/(),='
 
-def parse(statement):
-  return and_exp(list(Tokens(statement)))
+parse_options = xlocal()
+
+def get_attr(name):
+  return lambda row, ctx: getattr(row,name,None)
+
+def parse(statement, get_value=get_attr):
+  with parse_options(get_value=get_value):
+    return and_exp(list(Tokens(statement)))
 
 def and_exp(tokens):
+
   def and_(a, b):
-    def _and(row):
-      return a(row) and b(row)
+    def _and(row, ctx):
+      return a(row,ctx) and b(row,ctx)
     return _and
     
   lhs = or_exp(tokens) 
@@ -282,8 +324,8 @@ def and_exp(tokens):
   
 def or_exp(tokens):
   def or_(a, b):
-    def _or(row):
-      return a(row) or b(row)
+    def _or(row,ctx):
+      return a(row,ctx) or b(row,ctx)
     return _or
     
   lhs = comparison_exp(tokens)
@@ -298,8 +340,8 @@ def comparison_exp(tokens):
   if len(tokens) and tokens[0] in COMPARISON_OPS:
     op = COMPARISON_OPS[tokens.pop(0)]
     rhs =  additive_exp(tokens)
-    def comparison(row):
-      return op(lhs(row), rhs(row))
+    def comparison(row,ctx):
+      return op(lhs(row,ctx), rhs(row,ctx))
     return comparison
   else:
     return lhs
@@ -311,8 +353,8 @@ def additive_exp(tokens):
     if op:
       tokens.pop(0)
       rhs =  multiplicative_exp(tokens)
-      def additive(row, lhs=lhs): # bind lhs for each return
-        return op(lhs(row), rhs(row))
+      def additive(row, ctx, lhs=lhs): # bind lhs for each return
+        return op(lhs(row,ctx), rhs(row,ctx))
       lhs = additive
     else:
       break
@@ -325,8 +367,8 @@ def multiplicative_exp(tokens):
     if op:
       tokens.pop(0)
       rhs = unary_exp(tokens)
-      def multiplicative(row, lhs=lhs):
-        return op(lhs(row), rhs(row))
+      def multiplicative(row, ctx, lhs=lhs):
+        return op(lhs(row,ctx), rhs(row,ctx))
       lhs = multiplicative
     else:
       break
@@ -338,11 +380,11 @@ def unary_exp(tokens):
   if tokens[0] == '-':
     tokens.pop(0)
     value = value_exp(tokens)
-    return lambda row: operator.neg(value(row))
+    return lambda row, ctx: operator.neg(value(row, ctx))
   elif tokens[0] == 'not':
     tokens.pop(0)
     value = value_exp(tokens)
-    return lambda row: operator.not_(value(row))
+    return lambda row, ctx: operator.not_(value(row, ctx))
   elif tokens[0] == '+':
     tokens.pop(0)
     
@@ -355,19 +397,37 @@ def value_exp(tokens):
   token = tokens.pop(0)
   
   if token.startswith('$'):
-    return operator.itemgetter(int(token[1:]))
+    key = token[1:]
+    try:
+      key = int(key)
+    except ValueError:
+      pass
+    return operator.itemgetter(key)
+
+
+  if token.startswith('?'):
+    pos = int(token[1:])
+    return lambda row, ctx: ctx.get('params',[0])[pos]
+
   elif token[0] in string.digits:
-    return lambda x: int(token)
+    return lambda row, ctx: int(token)
   elif token.startswith('"'):
-    return lambda x: token[1:-1]
+    return lambda row, ctx: token[1:-1]
+  elif token in SYMBOLS: 
+    return lambda row, ctx: token
   elif token == '(':
     return group_exp(tokens)
   else:
+
     if tokens and tokens[0] == '(':
       return function_exp(token, tokens)
     else:
-      return operator.attrgetter(token)
 
+      attr = parse_options.get_value(token)
+
+      #attr = lambda row, ctx: get_value#get_value(row)# getattr(row, token)#row.get(token)
+      attr.__name__ = str(token)
+      return attr 
   
 def function_exp(name, tokens):
   token = tokens.pop(0)
@@ -385,7 +445,13 @@ def function_exp(name, tokens):
   assert tokens[0] == ')'
   tokens.pop(0)
 
-  return lambda x: datetime.now()
+  def invoke_udf(record, ctx):
+    udf = ctx['udf'][name]
+
+    t = [a(record, ctx) for a in args]
+    return udf(*t)
+  invoke_udf.__name__ = str(name)
+  return invoke_udf
 
 
 class Tokens(object):
@@ -421,10 +487,10 @@ class Tokens(object):
     return self.current_char == ''
     
   def is_letter(self):
-    return self.current_char in string.letters
+    return self.current_char in string.letters + '_'
     
   def is_number(self):
-    return self.current_char in string.digits + '_'
+    return self.current_char in string.digits 
   
   def is_letter_or_digit(self):
     return self.is_letter() or self.is_number()
@@ -444,10 +510,15 @@ class Tokens(object):
     return ''.join(buff)
     
   def read_symbol(self):
-    if self.current_char in '+-*/(),=':
+    if self.current_char in SYMBOLS:
       char = self.current_char
       self.read_char()
       return char
+    elif self.current_char in "?$":
+      char = self.current_char
+      self.read_char()
+      return char+self.read_word()
+
     elif self.current_char == '<':
       self.read_char()
       if self.current_char == '=':
